/*
 * $Id: cap2bmp.c 1.3 1996/11/23 23:05:44 Detlef Rel $
 *
 *
 * This program converts the HP38G/HP48 output of a screen-dump-to-printer
 * (ON-1) into a BMP file. A HP38G will always dump to the serial port, on
 * a HP48 you'll have to set flag -34 (Printing Device = serial) and open
 * the serial port manualy by doing an OPENIO or by the sequence RECV ON.
 *
 * If it's atarted w/ one argument, the arg is treated as the name of the
 * file to read the dump from, else the data is expected from stdin. The
 * resulting BMP file is written to stdout. A typical commandline:
 *
 * cap2bmp x.cap >test.bmp  which is the same as  cap2bmp <x.cap >test.bmp
 *
 * Note: due to MesS-DOS limitations, the input file can't be > ~32k, but
 * this is unlikely anyway :-)
 *
 *
 * $Log: cap2bmp.c $
 * Revision 1.3  1996/11/23 23:05:44  Detlef
 * Reversed output BMP, added some comments and error checks.
 *
 * Revision 1.2  1996/11/23 01:50:47  Detlef
 * 1st working version (sort off...) !
 *
 * (c) Copyright 1996, detlef@dmhh.hanse.de, all rights reserved.
 */


/*
 *	Includes
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	<io.h>			/* MesS-DOS */
#include	<fcntl.h>		/* MesS-DOS */

/*
 *	Defines
 */

#define	BI_RGB		0L
#define	BI_RLE8		1L
#define	BI_RLE4		2L

#define	ESC		0x1B

#define	BUFLEN		(32U * 1024)

/*
 *	Typedefinitions
 */

typedef unsigned char
    BYTE ;

typedef unsigned short
    WORD ;

typedef unsigned long
    DWORD ;

typedef	struct tagBITMAPFILEHEADER
    {
	WORD	bfType ;	/* Must be 'BM'				    */
	DWORD	bfSize ;	/* Size in BYTES of the file - must be '6'  */
	WORD	bfReserved1 ;	/* Reserved, must be set to 0		    */
	WORD	bfReserved2 ;	/* Reserved, must be set to 0		    */
	DWORD	bfOffBits ;	/* Specifies in bytes the offset from the   */
				/* the start of the BITMAPFILEHEADER of the */
				/* actual BMP in the file		    */
    }
    BITMAPFILEHEADER ;

typedef	struct tagBITMAPINFOHEADER
    {
	DWORD	biSize ;	/* Number of bytes required by the	    */
				/* BITMAPINFOHEADER structure		    */
	DWORD	biWidth ;	/* Width of the BMP in pixels		    */
	DWORD	biHeight ;	/* Height of the BMP in pixels		    */
	WORD	biPlanes ;	/* Number of planes for the target device,  */
				/* must be set to 1			    */
	WORD	biBitCount ;	/* Number of bits/pixel (1,4,8,24)	    */
	DWORD	biCompression ;	/* Type of compression for a compressed BMP.*/
				/* One of the following:		    */
				/* BI_RGB  : BMP not compressed		    */
				/* BI_RLE8 : Run-length encoded format for  */
				/*   BMP's with 8 bits/pixel. The compression*/
				/*   format is a two-byte format consisting */
				/*   of a count byte followed by a byte	    */
				/*   containing a color index.		    */
				/* BI_RLE4 : Run-length encoded format for  */
				/*   BMP's with 4 bits/pixel. The compression*/
				/*   format is a two-byte format consisting */
				/*   of a count byte followed by two	    */
				/*   word-lenght color indexes.		    */
	DWORD	biSizeImage ;	/* Size of image in bytes		    */
	DWORD	biXPelsPerMeter ; /* Horizontal resolution in pixel/meter   */
				  /* of the target device for the BMP	    */
	DWORD	biYPelsPerMeter ; /* Vertical resolution in pixel/meter     */
				  /* of the target device for the BMP	    */
	DWORD	biClrUsed ;	/* Number of color indexes in the color	    */
				/* table actually used by the BMP. If this  */
				/* value is 0, the BMP uses the max. number */
				/* of colors corresponding to the value of  */
				/* the biBitCount field.		    */
	DWORD	biClrImportant ; /* Number of color indexes that are	    */
				 /* considered important for displaying the */
				 /* BMP. All color are important on 0.	    */
    }
    BITMAPINFOHEADER ;

typedef	struct tagRGBQUAD
    {
	BYTE	rgbBlue ;
	BYTE	rgbGreen ;
	BYTE	rgbRed ;
	BYTE	rgbReserved ;
    }
    RGBQUAD ;

typedef	struct tagBITMAPINFO
    {
	BITMAPINFOHEADER	bmiHeader ;	/* Information about the BMP*/
	RGBQUAD			bmiColors[2] ;	/* Color look up table	    */
    }
    BITMAPINFO ;

/*
 *	Global data
 */

static BITMAPFILEHEADER
    BMPhead =
    {
	('M' << 8) | 'B',
	'6', 0, 0,
	sizeof( BITMAPFILEHEADER ) + sizeof( BITMAPINFO )
    } ;

static BITMAPINFO
    BMPinfo =
    {
	{
	    sizeof( BITMAPINFOHEADER ),
	    0,			/* Width of the BMP in pixels		    */
	    0,			/* Height of the BMP in pixels		    */
	    1, 1, BI_RGB,
	    0,			/* Size of image in bytes		    */
	    0,			/* Horizontal resolution in pixel/meter     */
	    0,			/* Vertical resolution in pixel/meter       */
	    2, 0
	},
	{
	    { 204, 204, 204, 0 },
	    {   0,   0,   0, 0 }
	}
    } ;

/*
 *	Functions
 */

void  main ( int argc, char *argv[] )
{

    FILE
	*fp ;				/* File pointer for reading data    */
    BYTE
	pat,
	*datb,				/* Data buffer *		    */
	*bmpb,				/* BMP buffer *			    */
	*dp,				/* pointer into datb *		    */
	*bp ;				/* pointer into bmpb *		    */
    int
	ch ;				/* Temporary character		    */
    WORD
	i, j, k, l,			/* Loop counter/temporay vars	    */
	dsz,				/* # of bytes in datb		    */
	w,				/* Width of a line in bytes	    */
	width,				/* BMP width			    */
	height ;			/* BMP height			    */

    fprintf( stderr, "\n" ) ;

    switch ( argc )			/* Parse command line		    */
    {
	case 1 :			/* We'll have to read from stdin    */
	    fp = stdin ;

	    setmode( fileno( stdin ), O_BINARY ) ;	/* MesS-DOS */

	    break ;

	case 2 :			/* Program started with one argument,*/
					/* assume it is a filename	     */
	    if ( ! (fp = fopen( argv[1], "rb" )) )
	    {
		fprintf( stderr,
			 "%s: unable to open file <%s>\n", argv[0], argv[1] ) ;
		exit( 1 ) ;
	    }

	    break ;

	default :			/* Unable to parse command line	    */
	    fprintf( stderr,
		     "$Id: cap2bmp.c 1.3 1996/11/23 23:05:44 Detlef Rel $\n"
		     "(c) 1996, Detlef Mueller, detlef@dmhh.hanse.de. "
		     "All rights reserved.\n\n"
		     "\nUsage: %s [file]\n",
		     argv[0], argv[0] ) ;
	    exit( 2 ) ;
    }

    if ( ! (datb = malloc( BUFLEN )) )	/* Alloc the data buffer	    */
	goto memerr ;

    dsz    =
    width  =
    height = 0 ;

    {					/* Read graphics data into datb	    */
	int
	    state = 0 ;

	for ( dp = datb ; (ch = fgetc( fp )) != EOF && dsz < BUFLEN ; )
	    switch ( state )
	    {
		case 0 :		/* Idle				    */
		    if ( ch == ESC )
			state = -1 ;

		    break ;

		case -1 :		/* Got ESC			    */
		    if ( ch > 166 )
			state = 0 ;	/* Unsupported ESC sequence, skip it*/
		    else
		    {
			state = ch ;	/* Data length			    */
			*dp++ = ch ;
			++dsz ;

			if ( ch > width )
			    width = ch ;/* Correct width		    */

			height += 8 ;	/* Correct height		    */
		    }

		    break ;

		default :		/* In line			    */
		    --state ;

		    *dp++ = ch ;
		    ++dsz ;
	    }

	if ( ! dsz )			/* Nothing found		    */
	{
	    free( datb ) ;
	    fprintf( stderr, "%s: no data found\n", argv[0] ) ;
	    exit( 3 ) ;
	}

	if ( dsz == BUFLEN )		/* Input too big		    */
	{
	    free( datb ) ;
	    fprintf( stderr, "%s: input > %u bytes\n", argv[0], BUFLEN ) ;
	    exit( 4 ) ;
	}

	if ( state > 0 )		/* Was still in line		    */
	{
	    free( datb ) ;
	    fprintf( stderr, "%s: input incomlete\n", argv[0] ) ;
	    exit( 5 ) ;
	}
    }
						/* Get bytes/line and correct*/
    w = (((width + 7) >> 3) + 3) & 0xFFFC ;	/* it to LWORD boundary	     */

    if ( ! (bmpb = calloc( 1, w * height )) )	/* Alloc pixel data buffer  */
    {
	free( datb ) ;
memerr:
	fprintf( stderr, "%s: memory allocation problem\n", argv[0] ) ;
	exit( 6 ) ;
    }

    bp = &bmpb[w * height] ;		/* Beyond the BMP buffer	    */
    dp = datb ;

    do					/* Convert dump to BMP format	    */
	for ( bp -= w << 3, l = *dp++, i = 0 ; i < l ; ++i )
	    for ( k = 1 << (7 - (i & 0x07)), pat = *dp++, j = 8 ; j-- ; pat >>= 1 )
		if ( pat & 0x01 )
		    bp[j * w + (i >> 3)] |= k ;
    while ( dp - datb < dsz ) ;

    free( datb ) ;

    BMPinfo.bmiHeader.biWidth  = width ;
    BMPinfo.bmiHeader.biHeight = height ;

    setmode( fileno( stdout ), O_BINARY ) ;		/* MesS-DOS */

    if (    fwrite( &BMPhead, sizeof( BITMAPFILEHEADER ), 1, stdout ) != 1
	 || fwrite( &BMPinfo, sizeof( BITMAPINFO ), 1, stdout ) != 1
	 || fwrite( bmpb, w * height, 1, stdout ) != 1 )
    {
	free( bmpb ) ;

	fprintf( stderr, "%s: write error\n", argv[0] ) ;
	exit( 7 ) ;
    }

    free( bmpb ) ;

    fprintf( stderr, "%s: Done\n", argv[0] ) ;

    exit( 0 ) ;
}
